@@ -56,7 +56,7 @@
 
 gboolean on_delete_event (GtkWidget *widget, GdkEvent *event, gpointer user_data);
 
-static void reread_cd (gboolean ignore_no_media);
+static void reread_cd (void);
 static void update_ui_for_album (AlbumDetails *album);
 static gboolean get_action_state_bool (const char *name);
 static void set_action_state (const char *detailed_name);
@@ -82,7 +82,7 @@ static GtkWidget *title_entry, *artist_entry, *composer_label, *composer_entry,
 static GtkWidget *entry_table; /* GtkTable containing composer_entry */
 static GtkTreeViewColumn *composer_column; /* Treeview column containing composers */
 static GtkWidget *track_listview, *extract_button, *play_button, *select_button;
-static GtkWidget *status_bar;
+static GtkWidget *multiple_album_dialog, *status_bar, *submit_button, *reload_button;
 GtkListStore *track_store;
 GtkCellRenderer *toggle_renderer, *title_renderer, *artist_renderer, *composer_renderer;
 
@@ -95,10 +95,12 @@ gboolean open_finished;
 gboolean extracting = FALSE;
 static gboolean duplication_enabled;
 
+static gboolean eject_activated;
 static gint total_no_of_tracks;
 static gint no_of_tracks_selected;
 static AlbumDetails *current_album;
 static char *current_submit_url = NULL;
+static GCancellable *reread_cancellable; /* weak reference */
 
 gboolean autostart = FALSE, autoplay = FALSE;
 static char *device = NULL, **uris = NULL;
@@ -190,14 +192,37 @@ static void on_quit_activate (GSimpleAction *action, GVariant *parameter, gpoint
   }
 }
 
+enum {
+  SJ_RESPONSE_EJECTED = 1
+};
+
+static void
+disc_ejected_cb (void)
+{
+  /* first make sure we're not playing */
+  stop_playback ();
+  stop_ui_hack ();
+  update_ui_for_album (NULL);
+  if (multiple_album_dialog != NULL &&
+      gtk_widget_get_mapped (multiple_album_dialog))
+    gtk_dialog_response (GTK_DIALOG (multiple_album_dialog),
+                         SJ_RESPONSE_EJECTED);
+  if (reread_cancellable != NULL) {
+    g_cancellable_cancel (reread_cancellable);
+    reread_cancellable = NULL;
+  }
+  set_action_state ("re-read(false)");
+  set_action_enabled ("re-read", FALSE);
+  set_action_enabled ("submit-tracks", FALSE);
+}
+
 /**
  * Clicked Eject
  */
 static void on_eject_activate (GSimpleAction *action, GVariant *parameter, gpointer data)
 {
-  /* first make sure we're not playing */
-  stop_playback ();
-  set_action_state ("re-read(false)");
+  disc_ejected_cb ();
+  eject_activated = TRUE;
   brasero_drive_eject (drive, FALSE, NULL);
 }
 
@@ -336,13 +361,15 @@ set_submit_text (const AlbumDetails *album)
   gchar *text;
 
   if (g_str_equal (album->title, _("Unknown Title"))) {
-    text = g_strdup (_("This album is not in the MusicBrainz database."));
+    text = g_strdup (_("This album is not in the MusicBrainz database, please click ‘Edit Album’ to open your brower and edit it in MusicBrainz."));
   } else {
-    text = g_strdup_printf (_("Could not find %s by %s on MusicBrainz."),
+    text = g_strdup_printf (_("Could not find %s by %s on MusicBrainz, please click ‘Edit Album’ to open your brower and edit it in MusicBrainz."),
                                     album->title, album->artist);
   }
   gtk_label_set_text (GTK_LABEL (submit_label), text);
   g_free (text);
+  gtk_widget_show (submit_button);
+  gtk_widget_hide (reload_button);
 }
 
 static void
@@ -355,12 +382,27 @@ submit_tracks_enabled_changed_cb (GActionGroup *action_group,
     gtk_widget_hide (submit_bar);
 }
 
+enum {
+  SJ_RESPONSE_SUBMIT = 1,
+  SJ_RESPONSE_REREAD
+};
+
 G_MODULE_EXPORT void
 submit_bar_response_cb (GtkInfoBar *infobar,
                         int         response_id,
                         gpointer    user_data)
 {
-  gtk_widget_hide (GTK_WIDGET (infobar));
+  if (!gtk_widget_get_visible (submit_bar))
+    return;
+
+  if (response_id == SJ_RESPONSE_SUBMIT) {
+    gtk_widget_hide (submit_button);
+    gtk_widget_show (reload_button);
+    gtk_label_set_text (GTK_LABEL (submit_label),
+                        _("Click ‘Reload’ to load the edited album details from MusicBrainz"));
+  } else {
+    gtk_widget_hide (GTK_WIDGET (infobar));
+  }
 }
 
 /**
@@ -388,7 +430,7 @@ static void on_submit_activate (GSimpleAction *action, GVariant *parameter, gpoi
       g_error_free (error);
     }
   }
-  gtk_widget_hide (GTK_WIDGET (submit_bar));
+  gtk_info_bar_response (GTK_INFO_BAR (submit_bar), SJ_RESPONSE_SUBMIT);
 }
 
 static void on_preferences_activate (GSimpleAction *action, GVariant *parameter, gpointer data)
@@ -900,9 +942,10 @@ static char *format_release_details (AlbumDetails *album)
 /**
  * Utility function for when there are more than one albums available
  */
-AlbumDetails* multiple_album_dialog(GList *albums)
+static AlbumDetails*
+choose_album(GList *albums)
 {
-  static GtkWidget *dialog = NULL, *albums_listview;
+  static GtkWidget *albums_listview;
   static GtkListStore *albums_store;
   static GtkTreeSelection *selection;
   AlbumDetails *album;
@@ -916,13 +959,14 @@ AlbumDetails* multiple_album_dialog(GList *albums)
     COLUMN_COUNT
   };
 
-  if (dialog == NULL) {
+  if (multiple_album_dialog == NULL) {
     GtkTreeViewColumn *column = gtk_tree_view_column_new ();
     GtkCellRenderer *renderer = gtk_cell_renderer_text_new ();
 
-    dialog = GET_WIDGET ("multiple_dialog");
-    g_assert (dialog != NULL);
-    gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (main_window));
+    multiple_album_dialog = GET_WIDGET ("multiple_dialog");
+    g_assert (multiple_album_dialog != NULL);
+    gtk_window_set_transient_for (GTK_WINDOW (multiple_album_dialog),
+                                  GTK_WINDOW (main_window));
     albums_listview = GET_WIDGET ("albums_listview");
     ok_button       = GET_WIDGET ("ok_button");
 
@@ -932,7 +976,7 @@ AlbumDetails* multiple_album_dialog(GList *albums)
                                         "markup", COLUMN_RELEASE);
 
     g_signal_connect (albums_listview, "row-activated",
-                      G_CALLBACK (album_row_activated), dialog);
+                      G_CALLBACK (album_row_activated), multiple_album_dialog);
 
     albums_store = gtk_list_store_new (COLUMN_COUNT,
                                        G_TYPE_STRING,
@@ -987,11 +1031,12 @@ AlbumDetails* multiple_album_dialog(GList *albums)
     gtk_tree_selection_select_iter (selection, &iter);
   }
 
-  gtk_widget_show_all (dialog);
-  response = gtk_dialog_run (GTK_DIALOG (dialog));
-  gtk_widget_hide (dialog);
+  gtk_widget_show_all (multiple_album_dialog);
+  response = gtk_dialog_run (GTK_DIALOG (multiple_album_dialog));
+  gtk_widget_hide (multiple_album_dialog);
 
-  if (response == GTK_RESPONSE_DELETE_EVENT) {
+  if (response == GTK_RESPONSE_DELETE_EVENT ||
+      response == SJ_RESPONSE_EJECTED) {
     return NULL;
   }
 
@@ -1115,10 +1160,30 @@ static void audio_volume_changed_cb (GSettings *settings, gchar *key, gpointer u
 }
 
 static void
-metadata_cb (SjMetadataGetter *m, GList *albums, GError *error)
+metadata_cb (GObject      *source,
+             GAsyncResult *result,
+             gpointer      user_data)
 {
+  SjMetadataGetter *mdg = SJ_METADATA_GETTER (source);
+  GError *error = NULL;
+  GList *albums;
+  gchar *url;
+
+  albums = sj_metadata_getter_list_albums_finish (mdg,
+                                                  result,
+                                                  &url,
+                                                  &error);
+  /* Check if this request was cancelled */
+  if (g_error_matches (error, G_IO_ERROR, G_IO_ERROR_CANCELLED)) {
+    g_error_free (error);
+    return;
+  }
+
+  reread_cancellable = NULL;
+  set_action_enabled ("re-read", TRUE);
   set_action_state ("re-read(false)");
-  if (error && !(error->code == SJ_ERROR_CD_NO_MEDIA)) {
+
+  if (error != NULL) {
     gboolean realized;
     GtkWidget *dialog;
 
@@ -1134,36 +1199,31 @@ metadata_cb (SjMetadataGetter *m, GList *albums, GError *error)
     gtk_dialog_run (GTK_DIALOG (dialog));
     gtk_widget_destroy (dialog);
     update_ui_for_album (NULL);
+    g_error_free (error);
     return;
   }
 
   g_free (current_submit_url);
-  current_submit_url = sj_metadata_getter_get_submit_url (metadata);
+  current_submit_url = url;
   if (current_submit_url) {
     set_action_enabled ("submit-tracks", TRUE);
   }
-  set_action_enabled ("re-read", TRUE);
-
   /* Free old album details */
-  if (current_album != NULL) {
-    album_details_free (current_album);
-    current_album = NULL;
-  }
+  g_clear_pointer (&current_album, (GDestroyNotify) album_details_free);
   /* Set the new current album pointer */
-  if (albums == NULL) {
-    current_album = NULL;
-  } else if (g_list_next (albums)) {
-    current_album = multiple_album_dialog (albums);
-    /* Concentrate here. We remove the album we want from the list, and then
-       deep-free the list. */
-    albums = g_list_remove (albums, current_album);
-    g_list_free_full (albums, (GDestroyNotify)album_details_free);
-    albums = NULL;
-  } else {
-    current_album = albums->data;
-    /* current_album now owns ->data, so just free the list */
-    g_list_free (albums);
-    albums = NULL;
+  if (albums != NULL) {
+    if (albums->next != NULL) {
+      current_album = choose_album (albums);
+      /* Concentrate here. We remove the album we want from the list, and then
+         deep-free the list. */
+      albums = g_list_remove (albums, current_album);
+      g_list_free_full (albums, (GDestroyNotify)album_details_free);
+      albums = NULL;
+    } else {
+      current_album = albums->data;
+      /* current_album now owns ->data, so just free the list */
+      g_clear_pointer (&albums, (GDestroyNotify) g_list_free);
+    }
   }
   update_ui_for_album (current_album);
 
@@ -1198,11 +1258,8 @@ is_audio_cd (BraseroDrive *drive)
 /**
  * Utility function to reread a CD
  */
-static void reread_cd (gboolean ignore_no_media)
+static void reread_cd (void)
 {
-  /* TODO: remove ignore_no_media? */
-  GError *error = NULL;
-
   set_action_state ("re-read(true)");
   set_action_enabled ("re-read", FALSE);
 
@@ -1223,29 +1280,16 @@ static void reread_cd (gboolean ignore_no_media)
     return;
   }
 
-  sj_metadata_getter_list_albums (metadata, &error);
+  if (reread_cancellable != NULL)
+    g_cancellable_cancel (reread_cancellable);
 
-  if (error && !(error->code == SJ_ERROR_CD_NO_MEDIA && ignore_no_media)) {
-    gboolean realized;
-    GtkWidget *dialog;
-
-    realized = gtk_widget_get_realized (main_window);
-    dialog = gtk_message_dialog_new (realized ? GTK_WINDOW (main_window) : NULL, 0,
-                                     GTK_MESSAGE_ERROR,
-                                     GTK_BUTTONS_CLOSE,
-                                     "%s", _("Could not read the CD"));
-    gtk_message_dialog_format_secondary_markup (GTK_MESSAGE_DIALOG (dialog),
-                                                "%s\n%s: %s",
-                                                _("Sound Juicer could not read the track listing on this CD."),
-                                                _("Reason"),
-                                                error->message);
-
-    gtk_dialog_run (GTK_DIALOG (dialog));
-    gtk_widget_destroy (dialog);
-    g_error_free (error);
-    update_ui_for_album (NULL);
-    return;
-  }
+  reread_cancellable = g_cancellable_new ();
+  sj_metadata_getter_list_albums_async (metadata,
+                                        reread_cancellable,
+                                        metadata_cb,
+                                        NULL);
+  /* reread_cancellable holds a weak reference */
+  g_object_unref (reread_cancellable);
 }
 
 static void
@@ -1260,7 +1304,7 @@ media_added_cb (BraseroMediumMonitor	*drive,
   sj_debug (DEBUG_CD, "Media added to device %s\n", brasero_drive_get_device (brasero_medium_get_drive (medium)));
   /* Don't call re-read if metadata is already being retreived */
   if (!get_action_state_bool ("re-read"))
-    reread_cd (TRUE);
+    reread_cd ();
 }
 
 static void
@@ -1272,13 +1316,12 @@ media_removed_cb (BraseroMediumMonitor	*drive,
     /* FIXME: recover? */
   }
 
-  /* first make sure we're not playing */
-  stop_playback ();
+  if (!eject_activated)
+    disc_ejected_cb ();
+  else
+    eject_activated = FALSE;
 
   sj_debug (DEBUG_CD, "Media removed from device %s\n", brasero_drive_get_device (brasero_medium_get_drive (medium)));
-  stop_ui_hack ();
-  update_ui_for_album (NULL);
-  set_action_enabled ("submit-tracks", FALSE);
 }
 
 static void
@@ -1318,7 +1361,7 @@ set_drive_from_device (const char *device)
 }
 
 static void
-set_device (const char* device, gboolean ignore_no_media)
+set_device (const char* device)
 {
   gboolean tray_opened;
 
@@ -1358,7 +1401,7 @@ set_device (const char* device, gboolean ignore_no_media)
   if (drive != NULL) {
     tray_opened = brasero_drive_is_door_open (drive);
     if (tray_opened == FALSE) {
-      reread_cd (ignore_no_media);
+      reread_cd ();
     }
 
     /* Enable/disable the eject options based on wether the drive supports ejection */
@@ -1415,7 +1458,6 @@ static void device_changed_cb (GSettings *settings, gchar *key, gpointer user_da
 {
   const char *device;
   char *value;
-  gboolean ignore_no_media = GPOINTER_TO_INT (user_data);
   g_assert (strcmp (key, SJ_SETTINGS_DEVICE) == 0);
 
   value = g_settings_get_string (settings, key);
@@ -1439,7 +1481,7 @@ static void device_changed_cb (GSettings *settings, gchar *key, gpointer user_da
   } else {
     device = value;
   }
-  set_device (device, ignore_no_media);
+  set_device (device);
   g_free (value);
 }
 
@@ -1486,7 +1528,8 @@ static void profile_changed_cb (GSettings *settings, gchar *key, gpointer user_d
  */
 static void on_reread_activate (GSimpleAction *action, GVariant *parameter, gpointer data)
 {
-  reread_cd (FALSE);
+  gtk_info_bar_response (GTK_INFO_BAR (submit_bar), SJ_RESPONSE_REREAD);
+  reread_cd ();
 }
 
 /**
@@ -2085,7 +2128,6 @@ startup_cb (GApplication *app, gpointer user_data)
   brasero_media_library_start ();
 
   metadata = sj_metadata_getter_new ();
-  g_signal_connect (metadata, "metadata", G_CALLBACK (metadata_cb), NULL);
 
   sj_settings = g_settings_new ("org.gnome.sound-juicer");
 
@@ -2145,6 +2187,8 @@ startup_cb (GApplication *app, gpointer user_data)
   main_window           = GET_WIDGET ("main_window");
   submit_bar            = GET_WIDGET ("submit_bar");
   submit_label          = GET_WIDGET ("submit_label");
+  reload_button         = GET_WIDGET ("submit_bar_reload_button");
+  submit_button         = GET_WIDGET ("submit_bar_submit_button");
   title_entry           = GET_WIDGET ("title_entry");
   artist_entry          = GET_WIDGET ("artist_entry");
   composer_label        = GET_WIDGET ("composer_label");
@@ -2335,7 +2379,7 @@ startup_cb (GApplication *app, gpointer user_data)
     device_changed_cb (sj_settings, SJ_SETTINGS_DEVICE, NULL);
   } else {
     if (device)
-      set_device (device, TRUE);
+      set_device (device);
     else {
       char *d;
 
@@ -2347,7 +2391,7 @@ startup_cb (GApplication *app, gpointer user_data)
       len = strlen (d);
       if (d[len - 1] == '/')
           d [len - 1] = '\0';
-      set_device (d, TRUE);
+      set_device (d);
       g_free (d);
       } else {
         device_changed_cb (sj_settings, SJ_SETTINGS_DEVICE, NULL);
